"""Test the Renogy Pro BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.renogy_pro_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import DefGATTChar, MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 512  # ATT max is 512 bytes


def ref_value() -> BMSSample:
    """Return reference value for mock Renogy Pro BMS."""
    return {
        "battery_charging": False,
        "battery_level": 99.0,
        "cell_voltages": [3.3, 3.3, 3.3, 3.3],
        "cell_count": 4,
        "current": -1.2,
        "cycle_capacity": 2779.195,
        "cycle_charge": 208.962,
        "cycles": 6,
        "delta_voltage": 0.0,
        "design_capacity": 211,
        "power": -15.96,
        "problem": False,
        "problem_code": 0,
        "runtime": 626886,
        "temp_values": [27.3, 26.8, 27.5],
        "temp_sensors": 3,
        "temperature": 27.2,
        "voltage": 13.3,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "heater": False,
    }


BASE_VALUE_CMD: Final[bytes] = b"\x30\x03\x13\xb2\x00\x07\xa4\x8a"


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockRenogyProBleakClient(MockBleakClient):
    """Emulate a Renogy Pro BMS BleakClient."""

    RESP: dict[bytes, bytearray] = {
        b"\x30\x03\x13\x88\x00\x22\x45\x5c": bytearray(
            b"\x30\x03\x44\x00\x04\x00\x21\x00\x21\x00\x21\x00\x21\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x01\x11\x01"
            b"\x0c\x01\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x65\x4a"
        ),
        BASE_VALUE_CMD: bytearray(
            b"\x30\x03\x0e\xff\xf4\x00\x85\x00\x03\x30\x42\x00\x03\x3b\xda\x00\x06\x3e\x33"
        ),  # -1.2A, 13.3V, 208.9Ah [mAh], 211.9Ah [mAh], 6 cycles
         b"\x30\x03\x13\xec\x00\x08\x85\x5c": bytearray(
            b"\x30\x03\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0e\x00\x00\xf7\x62"
        ),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, cmd: bytes
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffd1"):
            return bytearray()

        return self.RESP.get(cmd, bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        resp: bytearray = self._response(char_specifier, bytes(data))
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockRenogyProBleakClient", notify_data)

    @property
    def services(self) -> BleakGATTServiceCollection:
        """Emulate Renogy BT service setup."""

        serv_col = BleakGATTServiceCollection()
        for service in (
            BleakGATTService(None, 1, uuid=normalize_uuid_str("ffd0")),
            BleakGATTService(None, 5, uuid=normalize_uuid_str("fff0")),
        ):
            service.add_characteristic(
                DefGATTChar(
                    service.handle + 1,
                    uuid=f"{service.uuid[4:7]!s}1",
                    properties=["notify"],
                    service=service,
                )
            )
            service.add_characteristic(
                DefGATTChar(
                    service.handle + 2,
                    uuid=f"{service.uuid[4:7]!s}1",
                    properties=["write", "write-without-response"],
                    service=service,
                )
            )
            service.add_characteristic(
                DefGATTChar(
                    service.handle + 3,
                    uuid="0000",
                    properties=["write", "write-without-response"],
                    service=service,
                )
            )

            serv_col.add_service(service)

        return serv_col


class MockWrongBleakClient(MockBleakClient):
    """Mock client with invalid service for Renogy BMS."""

    @property
    def services(self) -> BleakGATTServiceCollection:
        """Emulate Renogy BT service setup."""

        return BleakGATTServiceCollection()


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Renogy Pro BMS data update."""

    patch_bleak_client(MockRenogyProBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_invalid_device(patch_bleak_client) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bleak_client(MockWrongBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}

    with pytest.raises(
        ConnectionError, match=r"^Failed to detect characteristics from.*"
    ):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()



"""Test the Lithionics BMS implementation."""

import asyncio
from collections.abc import Awaitable, Callable
import contextlib
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.lithionics_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 20

STREAM_DATA: Final[bytes] = (
    b"ERROR\r\n"
    b"1399,350,350,350,349,55,48,-3,99,000000\r\n"
    b"&,1,319,006391,0136,2300,FF05,8700\r\n"
)


def ref_value() -> BMSSample:
    """Return reference value for mock Lithionics BMS."""
    return {
        "voltage": 13.99,
        "current": -3.0,
        "battery_level": 99,
        "problem_code": 0,
        "temp_sensors": 2,
        "cell_count": 4,
        "cell_voltages": [3.5, 3.5, 3.5, 3.49],
        "temp_values": [12.778, 8.889],
        "temperature": 10.834,
        "cycle_charge": 319.0,
        "total_charge": 6391,
        "cycle_capacity": 4462.81,
        "delta_voltage": 0.01,
        "power": -41.97,
        "battery_charging": False,
        "runtime": 382800,
        "problem": False,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockLithionicsBleakClient(MockBleakClient):
    """Emulate a Lithionics BMS BleakClient."""

    _RESP: bytes = STREAM_DATA
    _task: asyncio.Task[None] | None = None

    async def _notify(self) -> None:
        """Notify function."""
        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        while True:
            for notify_data in [
                self._RESP[i : i + BT_FRAME_SIZE]
                for i in range(0, len(self._RESP), BT_FRAME_SIZE)
            ]:
                self._notify_callback(
                    "MockLithionicsBleakClient", bytearray(notify_data)
                )
            await asyncio.sleep(0.01)

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs,
    ) -> None:
        """Mock start_notify."""
        await super().start_notify(char_specifier, callback)
        self._task = asyncio.create_task(self._notify())

    async def disconnect(self) -> None:
        """Mock disconnect and wait for send task."""
        if self._task is not None:
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task
        await super().disconnect()


async def test_update(
    monkeypatch, patch_bleak_client, keep_alive_fixture: bool
) -> None:
    """Test Lithionics BMS data update."""
    monkeypatch.setattr(MockLithionicsBleakClient, "_RESP", STREAM_DATA)
    patch_bleak_client(MockLithionicsBleakClient)

    bms = BMS(generate_ble_device(name="Lithionics"), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockLithionicsBleakClient, BMS)


@pytest.fixture(
    name="invalid_stream",
    params=[
        b"",
        b"\r\n",
        b"ERROR\r\n",
        b"1,2,3\r\n",
        b"&,\r\n",
        b"&,1,2\r\n",
        b"text\r\n",
    ],
    ids=[
        "empty",
        "blank_line",
        "error_only",
        "short_primary",
        "short_status",
        "status_only",
        "unknown_line",
    ],
)
def fixture_invalid_stream(request: pytest.FixtureRequest) -> bytes:
    """Return invalid stream data payload."""
    assert isinstance(request.param, bytes)
    return request.param


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    invalid_stream: bytes,
) -> None:
    """Test data update with invalid stream data."""
    patch_bms_timeout("lithionics_bms")
    monkeypatch.setattr(MockLithionicsBleakClient, "_RESP", invalid_stream)
    patch_bleak_client(MockLithionicsBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


def test_uuid_tx_not_implemented() -> None:
    """Test that TX UUID is intentionally not implemented for stream-only protocol."""
    with pytest.raises(NotImplementedError):
        BMS.uuid_tx()


@pytest.mark.parametrize(
    ("status_line", "expected"),
    [
        ("&,", {}),
        ("&,1,2", {"cycle_charge": 2.0}),
    ],
    ids=["status_min_fields", "status_remaining_ah_only"],
)
async def test_status_field_variants(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    status_line: str,
    expected: BMSSample,
) -> None:
    """Test status parsing variants with optional fields."""
    stream = (
        b"1399,350,350,350,349,55,48,-3,99,000000\r\n" + status_line.encode() + b"\r\n"
    )
    monkeypatch.setattr(MockLithionicsBleakClient, "_RESP", stream)
    patch_bleak_client(MockLithionicsBleakClient)

    bms = BMS(generate_ble_device(name="Lithionics"))
    if expected:
        result = await bms.async_update()
    else:
        patch_bms_timeout("lithionics_bms")
        with pytest.raises(TimeoutError):
            await bms.async_update()
        await bms.disconnect()
        return

    for key, value in expected.items():
        assert result.get(key) == value

    await bms.disconnect()



"""Test the CBT power BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.cbtpwr_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


def ref_value() -> BMSSample:
    """Return reference value for mock CBT power BMS."""
    return {
        "voltage": 13.4,
        "current": -3.14,
        "battery_level": 100,
        "cycles": 3,
        "cycle_charge": 40.0,
        "cell_count": 5,
        "cell_voltages": [3.339, 3.339, 3.338, 3.338, 2.317],
        "delta_voltage": 1.022,
        "temp_values": [-2],
        "temperature": -2,
        "cycle_capacity": 536.0,
        "design_capacity": 40,
        "power": -42.076,
        "runtime": 22608,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockCBTpwrBleakClient(MockBleakClient):
    """Emulate a CBT power BMS BleakClient."""

    RESP: dict[int, bytes] = {
        0x05: b"\xaa\x55\x05\x0a\x0b\x0d\x0b\x0d\x0a\x0d\x0a\x0d\x0d\x09\x83\x0d\x0a",  # cell voltage info (5 cells)
        0x06: b"\xaa\x55\x06\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x0d\x0a",  # cell voltage info (no additional cells)
        0x09: b"\xaa\x55\x09\x0c\xfe\xff\xfe\xff\x00\x00\x00\x00\x00\x00\x00\x00\x0f\x0d\x0a",  # temperature frame
        0x0B: b"\xaa\x55\x0b\x08\x58\x34\x00\x00\xbc\xf3\xff\xff\x4c\x0d\x0a",  # voltage/current frame
        0x0A: b"\xaa\x55\x0a\x06\x64\x13\x0d\x00\x00\x00\x94\x0d\x0a",  # capacity frame
        0x0C: b"\xaa\x55\x0c\x0c\x00\x00\x00\x00\x5b\x06\x00\x00\x03\x00\x74\x02\xf2\x0d\x0a",  # runtime info frame, 6.28h*100
        0x15: b"\xaa\x55\x15\x04\x28\x00\x03\x00\x44\x0d\x0a",  # cycle info frame
        0x21: b"\xaa\x55\x21\x04\x00\x00\x00\x00\x25\x0d\x0a",  # warnings frame
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffe9"):
            return bytearray()
        cmd: int = int(bytearray(data)[2])
        assert bytearray(data)[4] == cmd, "incorrect CRC"
        if cmd in (0x07, 0x08):
            pytest.fail("only 5 cells available, do not query.")

        return bytearray(self.RESP.get(cmd, b""))

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        if (resp := self._response(char_specifier, data)) != bytearray():
            self._notify_callback("MockCBTpwrBleakClient", resp)


class MockAllCellsBleakClient(MockCBTpwrBleakClient):
    """Emulate a CBT power BMS BleakClient."""

    RESP: dict[int, bytes] = {
        0x05: b"\xaa\x55\x05\x0a\x0b\x0d\x0a\x0d\x09\x0d\x08\x0d\x07\x0d\x7d\x0d\x0a",
        0x06: b"\xaa\x55\x06\x0a\x06\x0d\x05\x0d\x04\x0d\x03\x0d\x02\x0d\x65\x0d\x0a",
        0x07: b"\xaa\x55\x07\x0a\x01\x0d\x00\x0d\xff\x0c\xfe\x0c\xfd\x0c\x4a\x0d\x0a",
        0x08: b"\xaa\x55\x08\x0a\xfc\x0c\xfb\x0c\xfa\x0c\xf9\x0c\xf8\x0c\x30\x0d\x0a",
        0x09: b"\xaa\x55\x09\x0c\x15\x00\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3f\x0d\x0a",  # temperature frame
        0x0B: b"\xaa\x55\x0b\x08\x58\x34\x00\x00\x00\x00\x00\x00\x9f\x0d\x0a",  # voltage/current frame
        0x15: b"\xaa\x55\x15\x04\x28\x00\x03\x00\x44\x0d\x0a",  # cycle info frame
        0x0A: b"\xaa\x55\x0a\x06\x64\x13\x0d\x00\x00\x00\x94\x0d\x0a",  # capacity frame
        0x0C: b"\xaa\x55\x0c\x0c\x00\x00\x00\x00\x5b\x06\x00\x00\x03\x00\x74\x02\xf2\x0d\x0a",  # runtime info frame, 6.28h*100
        0x21: b"\xaa\x55\x21\x04\x00\x00\x00\x00\x25\x0d\x0a",  # warnings frame
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ffe9"):
            return bytearray()
        cmd: int = int(bytearray(data)[2])

        return bytearray(self.RESP.get(cmd, b""))


async def test_update(
    patch_bleak_client, patch_bms_timeout, keep_alive_fixture: bool
) -> None:
    """Test CBT power BMS data update."""

    patch_bms_timeout()
    patch_bleak_client(MockCBTpwrBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.mark.parametrize(
    "wrong_response",
    [
        b"\x12\x34\x00\x00\x00\x56\x78",
        b"invalid_len",
        b"\xaa\x55\x15\x04\x00\x00\x00\x00\x00\x0d\x0a",
        b"",
    ],
    ids=[
        "invalid start/end",
        "invalid length",
        "wrong CRC",
        "empty frame",
    ],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockCBTpwrBleakClient,
        "RESP",
        MockCBTpwrBleakClient.RESP | {0x0B: bytearray(wrong_response)},
    )
    patch_bleak_client(MockCBTpwrBleakClient)

    bms = BMS(generate_ble_device())

    with pytest.raises(TimeoutError):
        await bms.async_update()

    await bms.disconnect()


async def test_partly_base_data(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockCBTpwrBleakClient,
        "RESP",
        MockCBTpwrBleakClient.RESP | {0x0B: MockCBTpwrBleakClient.RESP[0x15]},
    )  # wrong response to voltage/current req (0xB) with cycle info (0x15)
    patch_bleak_client(MockCBTpwrBleakClient)

    bms = BMS(generate_ble_device())

    with pytest.raises(ValueError, match="BMS data incomplete."):
        await bms.async_update()

    await bms.disconnect()


async def test_all_cell_voltages(patch_bleak_client, patch_bms_timeout) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()
    patch_bleak_client(MockAllCellsBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == {
        "voltage": 13.4,
        "current": 0,
        "battery_level": 100,
        "cycles": 3,
        "cycle_charge": 40.0,
        "cell_count": 20,
        "cell_voltages": [
            3.339,
            3.338,
            3.337,
            3.336,
            3.335,
            3.334,
            3.333,
            3.332,
            3.331,
            3.330,
            3.329,
            3.328,
            3.327,
            3.326,
            3.325,
            3.324,
            3.323,
            3.322,
            3.321,
            3.320,
        ],
        "delta_voltage": 0.019,
        "temperature": 21,
        "temp_values": [21],
        "cycle_capacity": 536.0,
        "design_capacity": 40,
        "power": 0,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
    }

    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            {0x21: bytearray(b"\xaa\x55\x21\x04\x01\x00\x00\x00\x26\x0d\x0a")},
            "first_bit",
        ),
        (
            {0x21: bytearray(b"\xaa\x55\x21\x04\x00\x00\x00\x80\xa5\x0d\x0a")},
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[dict[int, bytearray], str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    problem_response: tuple[dict[int, bytearray], str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(  # patch response dictionary to only problem reports (no other data)
        MockCBTpwrBleakClient, "RESP", MockCBTpwrBleakClient.RESP | problem_response[0]
    )

    patch_bms_timeout()
    patch_bleak_client(MockCBTpwrBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == ref_value() | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 31),
    }

    await bms.disconnect()



"""Test the ECO-WORTHY implementation."""

import asyncio
from collections.abc import Awaitable, Callable
import contextlib
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ecoworthy_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

_PROTO_DEFS: Final[dict[int, dict[int, bytearray]]] = {
    0x1: {  # protocol version 1
        0xA1: bytearray(
            b"\xa1\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x44\x00\x18\x00\x48\x00\x64\x05"
            b"\x31\xff\x8e\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x02\x00\x00\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x21\x86"
        ),
        0xA2: bytearray(  # 4 cells, 3 temp sensors
            b"\xa2\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x56\x00\x04\x0c\xfb\x0c\xfd\x0c"
            b"\xfb\x0c\xfa\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x03\x00\xcd"
            b"\x00\xc0\x00\xbe\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\x97\x6a"
        ),
    },
    0x2: {  # protocol version 2, MAC in front
        0xA1: bytearray(
            b"\xe2\xe7\x79\x8f\x4c\x66\xa1\x00\x00\x08\x03\x44\x00\x08\x00\x62\x00\x64\x05\x30\xff"
            b"\xc4\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x02"
            b"\x00\x00\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x87"
        ),
        0xA2: bytearray(  # 4 cells, 2 temp sensors
            b"\xe2\xe7\x79\x8f\x4c\x66\xa2\x00\x00\x08\x03\x56\x00\x04\x0c\xfb\x0c\xfc\x0c\xfb\x0c"
            b"\xf5\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x02\x00\xdc\x00\xd6"
            b"\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\x30\x58"
        ),
    },
}


_RESULT_DEFS: Final[dict[int, BMSSample]] = {
    0x1: {
        "cell_count": 4,
        "temp_sensors": 3,
        "voltage": 13.29,
        "current": -1.14,
        "battery_level": 72,
        "battery_health": 100,
        "cycle_charge": 72.0,
        "design_capacity": 100,
        #        "cycles": 8,
        "temperature": 19.567,
        "cycle_capacity": 956.88,
        "power": -15.151,
        "battery_charging": False,
        "cell_voltages": [3.323, 3.325, 3.323, 3.322],
        "temp_values": [20.5, 19.2, 19.0],
        "delta_voltage": 0.003,
        "runtime": 227368,
        "problem": False,
        "problem_code": 0,
    },
    0x2: {
        "cell_count": 4,
        "temp_sensors": 2,
        "voltage": 13.28,
        "current": -6.0,
        "battery_level": 98,
        "battery_health": 100,
        "cycle_charge": 98.0,
        "design_capacity": 100,
        "temperature": 21.7,
        "cycle_capacity": 1301.44,
        "power": -79.68,
        "battery_charging": False,
        "cell_voltages": [3.323, 3.324, 3.323, 3.317],
        "temp_values": [22.0, 21.4],
        "delta_voltage": 0.007,
        "problem": False,
        "problem_code": 0,
        "runtime": 58799,
    },
}


@pytest.fixture(
    name="protocol_type",
    params=[0x1, 0x2],
)
def proto(request: pytest.FixtureRequest) -> int:
    """Protocol fixture."""
    assert isinstance(request.param, int)
    return request.param


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockECOWBleakClient(MockBleakClient):
    """Emulate a ECO-WORTHY BMS BleakClient."""

    CMDS: Final[dict[int, bytearray]] = {
        0xA1: bytearray(b"\x00\x01\x03\x00\x8c\x00\x00\x99\x42"),
        0xA2: bytearray(b"\x00\x01\x03\x00\x8d\x00\x00\x59\x13"),
    }
    RESP: Final[dict[int, bytearray]] = {}

    _task: asyncio.Task[None] | None = None

    async def _notify(self) -> None:
        """Notify function."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        while True:
            for msg in self.RESP.values():
                self._notify_callback("MockECOWBleakClient", msg)
                await asyncio.sleep(0.01)

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs,
    ) -> None:
        """Issue write command to GATT."""
        await super().start_notify(char_specifier, callback, **kwargs)

        self._task = asyncio.create_task(self._notify())

    async def disconnect(self) -> None:
        """Mock disconnect and wait for send task."""
        if self._task is not None:
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task
        await super().disconnect()


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: int,
    keep_alive_fixture: bool,
) -> None:
    """Test ECO-WORTHY BMS data update."""

    monkeypatch.setattr(MockECOWBleakClient, "RESP", _PROTO_DEFS[protocol_type])
    patch_bleak_client(MockECOWBleakClient)

    bms = BMS(generate_ble_device("e2:e7:79:8f:4c:66"), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockECOWBleakClient, BMS)


async def test_tx_notimplemented(patch_bleak_client) -> None:
    """Test ECO-WORTHY BMS uuid_tx not implemented for coverage."""

    patch_bleak_client(MockECOWBleakClient)

    bms = BMS(generate_ble_device(), False)

    with pytest.raises(NotImplementedError):
        _ret: str = bms.uuid_tx()


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            bytearray(
                b"\xa3\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x44\x00\x18\x00\x48\x00\x64"
                b"\x05\x31\xff\x8e\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x01\x00\x02\x00\x00\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x70\x20"
            ),
            "wrong_type",
        ),
        (
            bytearray(
                b"\xa2\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x56\x00\x04\x0c\xfb\x0c\xfd"
                b"\x0c\xfb\x0c\xfa\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
                b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
                b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
                b"\x00\x03\x00\xcd\x00\xc0\x00\xbe\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18\xfc\x18"
                b"\x97\x6a"
            ),
            "single_type_sent",
        ),
        (
            bytearray(  # correct CRC: 0x2186
                b"\xa1\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x44\x00\x18\x00\x48\x00\x64"
                b"\x05\x31\xff\x8e\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x01\x00\x02\x00\x00\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x21\x87"
            ),
            "wrong_CRC",
        ),
        (
            bytearray(b""),
            "empty_response",
        ),
    ],
    ids=lambda param: param[1],
)
def response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    protocol_type: int,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout("ecoworthy_bms")

    monkeypatch.setattr(MockECOWBleakClient, "RESP", _PROTO_DEFS[protocol_type])
    monkeypatch.setattr(
        MockECOWBleakClient,
        "RESP",
        {0xA1: wrong_response, 0xA2: MockECOWBleakClient.RESP[0xA2]},
    )

    patch_bleak_client(MockECOWBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\xa1\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x44\x00\x18\x00\x48\x00\x64"
                b"\x05\x31\xff\x8e\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x01\x00\x02\x00\x00\xff\xff\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x20\xd6"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\xa1\x00\x00\x00\x65\x00\x00\x00\x00\x00\x18\x01\x03\x44\x00\x18\x00\x48\x00\x64"
                b"\x05\x31\xff\x8e\x00\x00\x27\x10\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x01\x00\x02\x00\x00\xff\xff\x00\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x20\x0e"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(MockECOWBleakClient, "RESP", _PROTO_DEFS[0x1])
    monkeypatch.setattr(
        MockECOWBleakClient,
        "RESP",
        {
            0xA1: problem_response[0],
            0xA2: MockECOWBleakClient.RESP[0xA2],
        },
    )

    patch_bleak_client(MockECOWBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == _RESULT_DEFS[0x1] | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 15),
    }

    await bms.disconnect()



"""Test the Braun Power BMS implementation."""

import asyncio
from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.braunpwr_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


def ref_value() -> BMSSample:
    """Return reference value for mock CBT power BMS."""
    return {
        "voltage": 53.31,
        "current": -11.45,
        "battery_level": 83,
        "battery_health": 100,
        "cell_count": 16,
        "cycle_charge": 250.26,
        "cycles": 8,
        "cell_voltages": [
            3.327,
            3.334,
            3.333,
            3.335,
            3.333,
            3.335,
            3.333,
            3.338,
            3.332,
            3.335,
            3.334,
            3.336,
            3.333,
            3.334,
            3.336,
            3.337,
        ],
        "design_capacity": 300,
        "power": -610.399,
        "temp_sensors": 4,
        "temp_values": [
            23.0,
            23.0,
            23.0,
            23.0,
        ],
        "delta_voltage": 0.011,
        "cycle_capacity": 13341.361,
        "battery_charging": False,
        "temperature": 23.0,
        "runtime": 78684,
        "problem_code": 0,
        "problem": False,
        "balancer": 32385,
    }


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockBraunPWRBleakClient(MockBleakClient):
    """Emulate a Braun Power BMS BleakClient."""

    RESP: dict[int, bytearray] = {
        0x01: bytearray(
            b"\x7b\x01\x20\x00\x53\x14\xd3\x00\xd2\x00\xb4\x00\xbe\xfb\x87\x61"
            b"\xc2\x75\x30\x00\x00\x00\x00\x00\x08\x7e\x81\x00\x00\x00\x0e\x00"
            b"\x00\x00\x64\x7d"
        ),
        0x02: bytearray(
            b"\x7b\x02\x21\x10\x0c\xff\x0d\x06\x0d\x05\x0d\x07\x0d\x05\x0d\x07"
            b"\x0d\x05\x0d\x0a\x0d\x04\x0d\x07\x0d\x06\x0d\x08\x0d\x05\x0d\x06"
            b"\x0d\x08\x0d\x09\x7d"
        ),
        0x03: bytearray(b"{\x03\x09\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91}"),
        0x08: bytearray(
            b"\x7b\x08\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d"
        ),
        0x09: bytearray(b"\x7b\x09\x03\xff\xff\xff\x7d"),
        0x74: bytearray(
            b"\x7b\x74\x1d\x4b\x53\x5f\x42\x4c\x45\x5f\x57\x49\x46\x49\x5f\x56"
            b"\x65\x72\x31\x2e\x30\x2e\x30\x5f\x32\x30\x32\x34\x30\x33\x31\x33\x7d"
        ),
        0x78: bytearray(b"\x7b\x78\x02\x00\x23\x7d"),
        0xF4: bytearray(b"\x7b\xf4\x03\x02\x03\x01\x7d"),
        0xF5: bytearray(b"\x7b\xf5\x02\x02\x39\x7d"),
    }

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, cmd: int
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("ff02"):
            return bytearray()

        return self.RESP.get(cmd, bytearray())

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        self._notify_callback(
            "MockABCBleakClient", self._response(char_specifier, bytes(data)[1])
        )
        await asyncio.sleep(0)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test ABC BMS data update."""

    patch_bleak_client(MockBraunPWRBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockBraunPWRBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "hw_version": "KS_BLE_WIFI_Ver1.0.0_20240313",
        "sw_version": "2.3.1",
    }


@pytest.fixture(
    name="wrong_response",
    params=[
        # (
        #     b"!\x03\x09\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91}",
        #     "wrong_SOF",
        # ),
        (
            b"!\x03\x09\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91",
            "wrong_EOF",
        ),
        (
            b"{\x04\x09\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91}",
            "unknown_CMD",
        ),
        (
            b"{\x02\x09\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91}",
            "wrong_CMD",
        ),
        (
            b"{\x03\x08\x04\x0b\x91\x0b\x91\x0b\x91\x0b\x91}",
            "wrong_length",
        ),
    ],
    ids=lambda param: param[1],
)
def fix_response(request) -> bytearray:
    """Return faulty response frame."""
    return bytearray(request.param[0])


async def test_invalid_response(
    monkeypatch, patch_bleak_client, patch_bms_timeout, wrong_response: bytearray
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockBraunPWRBleakClient,
        "RESP",
        MockBraunPWRBleakClient.RESP | {0x03: wrong_response},
    )

    patch_bleak_client(MockBraunPWRBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            b"\x7b\x01\x20\x00\x53\x14\xd3\x00\xd2\x00\xb4\x00\xbe\xfb\x87\x61\xc2\x75\x30\x00\x00"
            b"\x00\x00\x00\x08\x00\x00\x00\x00\x00\x0e\x80\x00\x00\x64\x7d",
            0x8000,
            "first_bit",
        ),
        (
            b"\x7b\x01\x20\x00\x53\x14\xd3\x00\xd2\x00\xb4\x00\xbe\xfb\x87\x61\xc2\x75\x30\x00\x00"
            b"\x00\x00\x00\x08\x00\x00\x00\x00\x00\x0e\x00\x02\x00\x64\x7d",
            0x02,
            "last_bit",
        ),
        (
            b"\x7b\x01\x20\x00\x53\x14\xd3\x00\xd2\x00\xb4\x00\xbe\xfb\x87\x61\xc2\x75\x30\x00\x00"
            b"\x00\x00\x00\x08\x00\x00\x00\x00\x00\x0e\xff\xfe\x00\x64\x7d",
            0xFFFE,
            "all_bits_set",
        ),
    ],
    ids=lambda param: param[2],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, int, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, int, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockBraunPWRBleakClient,
        "RESP",
        MockBraunPWRBleakClient.RESP | {0x1: bytearray(problem_response[0])},
    )

    patch_bleak_client(MockBraunPWRBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result.get("problem", False)  # expect a problem report
    assert result.get("problem_code", 0) == problem_response[1]

    await bms.disconnect()



"""Test the TDT implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakDeviceNotFoundError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.tdt_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 27

_PROTO_DEFS: Final[dict[str, dict[int, bytearray]]] = {
    "4S4Tv0.0": {
        0x8C: bytearray(  # 4 cell message
            b"\x7e\x00\x01\x03\x00\x8c\x00\x20\x04\x0c\xe1\x0c\xdf\x0c\xe1\x0c"
            b"\xdc\x04\x0b\x93\x0b\x9b\x0b\x8d\x0b\x8c\x40\x00\x05\x26\x02\x3f"
            b"\x04\x1c\x00\x08\x03\xe8\x00\x37\x91\x91\x0d"
        ),
        0x8D: bytearray(
            b"\x7e\x00\x41\x03\x00\x8d\x00\x18\x04\x00\x00\x00\x00\x04\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x06\x09\x00\x00\x18\x00\x00\x00"
            b"\xdf\x68\x0d"
        ),
        0x95: bytearray(
            b"\x0e\x2c\x00\x03\xa1\xde\x06\x0a\x03\x20\x3d\x31\x00\x7d\xdf\xc5"
            b"\xa5\x6f\x9b\x9a\xaf\x0e\x19\x15\x00\x04\x00\x12\x14\x00\x7e\x00"
            b"\x01\x06\x00\x95\x00\x07\x07\xe8\x0b\x16\x0e\x07\x08\xc9\x41\x0d"
            b"\x79\xf7\x25"
        ),
    },
    "16S6Tv0.0": {
        0x8C: bytearray(  # 16 cell message
            b"\x7e\x00\x01\x03\x00\x8c\x00\x3c\x10\x0c\xe3\x0c\xe6\x0c\xde\x0c\xde\x0c\xdd\x0c"
            b"\xde\x0c\xdd\x0c\xdc\x0c\xdc\x0c\xda\x0c\xde\x0c\xde\x0c\xde\x0c\xdd\x0c\xdf\x0c"
            b"\xde\x06\x0b\x5e\x0b\x6f\x0b\x5e\x0b\x5e\x0b\x5e\x0b\x66\xc0\x39\x14\x96\x03\xdf"
            b"\x04\x3b\x00\x08\x03\xe8\x00\x5b\x2b\x9c\x0d"
        ),
        0x8D: bytearray(
            b"\x7e\x00\x01\x03\x00\x8d\x00\x27\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x0e\x01"
            b"\x00\x00\x18\x00\x00\x00\x00\x0b\x7b\x0d"
        ),
        0x92: bytearray(
            b"\x7e\x00\x01\x03\x00\x92\x00\x3c\x36\x30\x33\x32\x5f\x31\x30\x30\x31\x36\x53\x30"
            b"\x30\x30\x5f\x4c\x5f\x34\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x36\x30\x33\x32\x36\x30\x31\x36\x32\x30\x37\x32"
            b"\x37\x30\x30\x30\x31\x00\x00\x00\x50\x01\x0d"
        ),
    },
    "16S6Tv0.4": {  # version 0.4
        0x8C: bytearray(  # 16 cell message
            b"\x7e\x04\x01\x03\x00\x8c\x00\x3c\x10\x0c\xb2\x0c\xb8\x0c\xb5\x0c\xb8\x0c\xb4\x0c"
            b"\xb6\x0c\xb8\x0c\xb9\x0c\xb5\x0c\xb9\x0c\xb9\x0c\xb9\x0c\xba\x0c\xb7\x0c\xb7\x0c"
            b"\xb8\x06\x0b\xa9\x0b\xa6\x0b\xb8\x0b\xae\x0b\xc2\x0b\xbb\xc0\xa2\x14\x60\x02\x85"
            b"\x04\xfa\x00\xd5\x04\xb0\x00\x33\x2f\xb4\x0d"
        ),
        0x8D: bytearray(
            b"\x7e\x04\x01\x03\x00\x8d\x00\x27\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x0e\xfe"
            b"\x00\x00\x00\x00\x00\x00\x00\x3f\x56\x0d"
        ),
    },
}


def ref_value() -> dict[str, BMSSample]:
    """Return reference value for mock Seplos BMS."""
    return {
        "4S4Tv0.0": {
            "cell_count": 4,
            "temp_sensors": 4,
            "voltage": 13.18,
            "current": 0.0,
            "battery_level": 55,
            "cycle_charge": 57.5,
            "cycles": 8,
            "temperature": 23.075,
            "cycle_capacity": 757.85,
            "power": 0.0,
            "battery_charging": False,
            "cell_voltages": [3.297, 3.295, 3.297, 3.292],
            "temp_values": [23.2, 24.0, 22.6, 22.5],
            "delta_voltage": 0.005,
            "chrg_mosfet": True,
            "dischrg_mosfet": True,
            "problem": False,
            "problem_code": 0,
        },
        "16S6Tv0.0": {
            "cell_count": 16,
            "temp_sensors": 6,
            "voltage": 52.7,
            "current": -5.7,
            "battery_level": 91,
            "cycle_charge": 99.1,
            "cycles": 8,
            "temperature": 18.317,
            "cycle_capacity": 5222.57,
            "power": -300.39,
            "battery_charging": False,
            "cell_voltages": [
                3.299,
                3.302,
                3.294,
                3.294,
                3.293,
                3.294,
                3.293,
                3.292,
                3.292,
                3.290,
                3.294,
                3.294,
                3.294,
                3.293,
                3.295,
                3.294,
            ],
            "temp_values": [17.9, 19.6, 17.9, 17.9, 17.9, 18.7],
            "delta_voltage": 0.012,
            "runtime": 62589,
            "chrg_mosfet": True,
            "dischrg_mosfet": True,
            "problem": False,
            "problem_code": 0,
        },
        "16S6Tv0.4": {
            "cell_count": 16,
            "temp_sensors": 6,
            "cell_voltages": [
                3.25,
                3.256,
                3.253,
                3.256,
                3.252,
                3.254,
                3.256,
                3.257,
                3.253,
                3.257,
                3.257,
                3.257,
                3.258,
                3.255,
                3.255,
                3.256,
            ],
            "temp_values": [25.4, 25.1, 26.9, 25.9, 27.9, 27.2],
            "voltage": 52.16,
            "current": -16.2,
            "cycle_charge": 64.5,
            "battery_level": 51,
            "cycles": 213,
            "problem_code": 0,
            "delta_voltage": 0.008,
            "cycle_capacity": 3364.32,
            "power": -844.992,
            "battery_charging": False,
            "runtime": 14333,
            "temperature": 26.4,
            "chrg_mosfet": True,
            "dischrg_mosfet": True,
            "problem": False,
        },
    }


@pytest.fixture(
    name="protocol_type",
    params=["4S4Tv0.0", "16S6Tv0.0", "16S6Tv0.4"],
)
def proto(request: pytest.FixtureRequest) -> str:
    """Protocol fixture."""
    assert isinstance(request.param, str)
    return request.param


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockTDTBleakClient(MockBleakClient):
    """Emulate a TDT BMS BleakClient."""

    HEAD_CMD: Final[int] = 0x7E
    TAIL_CMD: Final[int] = 0x0D
    CMDS: Final[dict[int, bytearray]] = {
        0x8C: bytearray(b"\x00\x01\x03\x00\x8c\x00\x00"),
        0x8D: bytearray(b"\x00\x01\x03\x00\x8d\x00\x00"),
        0x92: bytearray(b"\x00\x01\x03\x00\x92\x00\x00"),
    }
    RESP: Final[dict[int, bytearray]] = {}

    _char_fffa: int = 0x0  # return value for UUID "fffa"

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("fff2")
            and bytearray(data)[0] == self.HEAD_CMD
            and bytearray(data)[-1] == self.TAIL_CMD
        ):
            for k, v in self.CMDS.items():
                if bytearray(data)[1:].startswith(v) and k in self.RESP:
                    return self.RESP[k]

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data)

        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fffa"):
            if data == b"HiLink":
                self._char_fffa = 0x1
            return

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: bytearray = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockTDTBleakClient", notify_data)

    async def read_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        **kwargs,
    ) -> bytearray:
        """Mock read GATT characteristics."""

        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier)[4:8] == "fffa"
        ):
            return bytearray(int.to_bytes(self._char_fffa, 1, "big"))

        return await super().read_gatt_char(char_specifier, kwargs=kwargs)


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: str,
    keep_alive_fixture: bool,
) -> None:
    """Test TDT BMS data update."""

    monkeypatch.setattr(MockTDTBleakClient, "RESP", _PROTO_DEFS[protocol_type])
    patch_bleak_client(MockTDTBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()[protocol_type]

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_0x1e_head(
    monkeypatch: pytest.MonkeyPatch,
    patch_bms_timeout,
    patch_bleak_client,
    keep_alive_fixture: bool,
) -> None:
    """Test TDT BMS data update."""

    resp_0x1e: Final[dict[int, bytearray]] = {
        0x8C: bytearray(  # 4 cell message
            b"\x7e\x00\x01\x03\x00\x8c\x00\x20\x04\x0c\xe1\x0c\xdf\x0c\xe1\x0c"
            b"\xdc\x04\x0b\x93\x0b\x9b\x0b\x8d\x0b\x8c\x40\x00\x05\x26\x02\x3f"
            b"\x04\x1c\x00\x08\x03\xe8\x00\x37\x91\x91\x0d"
        ),
        0x8D: bytearray(  # independent of the query, reply header is always 0x7E
            b"\x7e\x00\x41\x03\x00\x8d\x00\x18\x04\x00\x00\x00\x00\x04\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x06\x09\x00\x00\x18\x00\x00\x00"
            b"\xdf\x68\x0d"
        ),
    }

    patch_bms_timeout()
    monkeypatch.setattr(MockTDTBleakClient, "HEAD_CMD", 0x1E)
    monkeypatch.setattr(MockTDTBleakClient, "RESP", resp_0x1e)
    patch_bleak_client(MockTDTBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == ref_value()["4S4Tv0.0"]

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.mark.parametrize(("header"), [0x7E, 0x1E], ids=["7e", "1e"])
async def test_device_info(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    protocol_type: str,
    header: int,
) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    monkeypatch.setattr(MockTDTBleakClient, "RESP", _PROTO_DEFS[protocol_type])
    monkeypatch.setattr(MockTDTBleakClient, "HEAD_CMD", header)
    patch_bms_timeout()
    patch_bleak_client(MockTDTBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == (
        {
            "fw_version": "mock_FW_version",
            "hw_version": "mock_HW_version",
            "manufacturer": "mock_manufacturer",
            "model": "mock_model",
            "serial_number": "mock_serial_number",
            "sw_version": "mock_SW_version",
        }
        if protocol_type in ("4S4Tv0.0", "16S6Tv0.4")
        else {
            "sw_version": "6032_10016S000_L_41",
            "manufacturer": "",
            "serial_number": "60326016207270001",
        }
    )


@pytest.fixture(
    name="wrong_response",
    params=[
        (b"\x7e\x00\x01\x03\x00\x8c\x00\x01\x00\xa1\x18\x00", "invalid frame end"),
        (b"\x7e\x10\x01\x03\x00\x8c\x00\x01\x00\xad\x19\x0d", "invalid version"),
        (b"\x7e\x00\x01\x03\x00\x8c\x00\x01\x00\xa1\x00\x0d", "invalid CRC"),
        (b"\x7e\x00\x01\x03\x00\x8c\x00\x01\x00\xa1\x18\x0d\x00", "oversized frame"),
        (b"\x7e\x00\x01\x03\x00\x8c\x00\x01\x00\xa1\x0d", "undersized frame"),
        (b"\x7e\x00\x01\x03\x01\x8c\x00\x01\x00\x61\x25\x0d", "error response"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request):
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockTDTBleakClient, "_response", lambda _s, _c, _d: wrong_response
    )

    patch_bleak_client(MockTDTBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(("throw_exception"), [True, False], ids=("throw", "error"))
async def test_init_fail(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, throw_exception: bool
) -> None:
    """Test that failing to initialize simply continues and tries to read data."""

    async def error_repsonse(*_args, **_kwargs) -> bytearray:
        return bytearray(b"\x00")

    async def throw_response(*_args, **_kwargs) -> bytearray:
        raise BleakDeviceNotFoundError("MockTDTBleakClient")

    monkeypatch.setattr(MockTDTBleakClient, "RESP", _PROTO_DEFS["16S6Tv0.0"])
    monkeypatch.setattr(
        MockTDTBleakClient,
        "read_gatt_char",
        throw_response if throw_exception else error_repsonse,
    )

    patch_bleak_client(MockTDTBleakClient)

    bms = BMS(generate_ble_device())

    if throw_exception:
        with pytest.raises(BleakDeviceNotFoundError):
            assert not await bms.async_update()
    else:
        assert await bms.async_update() == ref_value()["16S6Tv0.0"]

    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            {
                0x8C: bytearray(  # 4 cell message
                    b"\x7e\x00\x01\x03\x00\x8c\x00\x20\x04\x0c\xe1\x0c\xdf\x0c\xe1\x0c"
                    b"\xdc\x04\x0b\x93\x0b\x9b\x0b\x8d\x0b\x8c\x40\x00\x05\x26\x02\x3f"
                    b"\x04\x1c\x00\x08\x03\xe8\x00\x37\x91\x91\x0d"
                ),
                0x8D: bytearray(
                    b"\x7e\x00\x41\x03\x00\x8d\x00\x18\x04\x00\x00\x00\x00\x04\x00\x00"
                    b"\x00\x00\x00\x00\x00\x00\x00\x01\x06\x09\x00\x00\x18\x00\x00\x00"
                    b"\x4f\x65\x0d"  #          ^^  ^^ problem bits
                ),
            },
            "first_bit_4cell",
        ),
        (
            {
                0x8C: bytearray(  # 4 cell message
                    b"\x7e\x00\x01\x03\x00\x8c\x00\x20\x04\x0c\xe1\x0c\xdf\x0c\xe1\x0c"
                    b"\xdc\x04\x0b\x93\x0b\x9b\x0b\x8d\x0b\x8c\x40\x00\x05\x26\x02\x3f"
                    b"\x04\x1c\x00\x08\x03\xe8\x00\x37\x91\x91\x0d"
                ),
                0x8D: bytearray(
                    b"\x7e\x00\x41\x03\x00\x8d\x00\x18\x04\x00\x00\x00\x00\x04\x00\x00"
                    b"\x00\x00\x00\x00\x00\x00\x80\x00\x06\x09\x00\x00\x18\x00\x00\x00"
                    b"\x37\x6f\x0d"  #          ^^  ^^ problem bits
                ),
            },
            "last_bit_4cell",
        ),
        (
            {
                0x8C: bytearray(  # 16 cell message
                    b"\x7e\x00\x01\x03\x00\x8c\x00\x3c\x10\x0c\xe3\x0c\xe6\x0c\xde\x0c\xde\x0c\xdd"
                    b"\x0c\xde\x0c\xdd\x0c\xdc\x0c\xdc\x0c\xda\x0c\xde\x0c\xde\x0c\xde\x0c\xdd\x0c"
                    b"\xdf\x0c\xde\x06\x0b\x5e\x0b\x6f\x0b\x5e\x0b\x5e\x0b\x5e\x0b\x66\xc0\x39\x14"
                    b"\x96\x03\xdf\x04\x3b\x00\x08\x03\xe8\x00\x5b\x2b\x9c\x0d"
                ),
                0x8D: bytearray(
                    b"\x7e\x00\x01\x03\x00\x8d\x00\x27\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                    b"\x00\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01"
                    b"\x0e\x01\x00\x00\x18\x00\x00\x00\x00\xce\x2a\x0d"  #     problem bits ^^  ^^
                ),
            },
            "first_bit_16cell",
        ),
        (
            {
                0x8C: bytearray(  # 16 cell message
                    b"\x7e\x00\x01\x03\x00\x8c\x00\x3c\x10\x0c\xe3\x0c\xe6\x0c\xde\x0c\xde\x0c\xdd"
                    b"\x0c\xde\x0c\xdd\x0c\xdc\x0c\xdc\x0c\xda\x0c\xde\x0c\xde\x0c\xde\x0c\xdd\x0c"
                    b"\xdf\x0c\xde\x06\x0b\x5e\x0b\x6f\x0b\x5e\x0b\x5e\x0b\x5e\x0b\x66\xc0\x39\x14"
                    b"\x96\x03\xdf\x04\x3b\x00\x08\x03\xe8\x00\x5b\x2b\x9c\x0d"
                ),
                0x8D: bytearray(
                    b"\x7e\x00\x01\x03\x00\x8d\x00\x27\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                    b"\x00\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x80\x00"
                    b"\x0e\x01\x00\x00\x18\x00\x00\x00\x00\xc9\xd2\x0d"  #     problem bits ^^  ^^
                ),
            },
            "last_bit_16cell",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(
    request: pytest.FixtureRequest,
) -> tuple[dict[int, bytearray], str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, problem_response: tuple[dict[int, bytearray], str]
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(MockTDTBleakClient, "RESP", problem_response[0])
    patch_bleak_client(MockTDTBleakClient)

    bms = BMS(generate_ble_device(), False)

    result: BMSSample = await bms.async_update()
    assert result.get("problem", False)  # we expect a problem
    assert result.get("problem_code", 0) == (
        0x1 if problem_response[1].startswith("first_bit") else 0x8000
    )

    await bms.disconnect()



"""Module to support Dummy BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Dummy BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Dummy Manufacturer",
        "default_model": "dummy model",
    }  # TODO: fill correct manufacturer/model
    # _HEAD: Final[bytes] = b"\x55"  # beginning of frame
    # _TAIL: Final[bytes] = b"\xAA"  # end of frame
    # _FRAME_LEN: Final[int] = 10  # length of frame, including SOF and checksum

    # accept_secret: bool = True  # if the BMS accepts a secret for authentication

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"local_name": "dummy", "connectable": True}]  # TODO: define matcher

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("0000"),)  # TODO: change service UUID here!

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "0000"  # TODO: change RX characteristic UUID here!

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "0000"  # TODO: change TX characteristic UUID here!

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        return BMSInfo(
            default_manufacturer="Dummy manufacturer", default_model="Dummy BMS"
        )  # TODO: implement query code or remove function to query service 0x180A

    # @staticmethod
    # def _raw_values() -> frozenset[BMSValue]:
    #     return frozenset({"runtime"})  # never calculate, e.g. runtime

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        # self._log.debug("RX BLE data: %s", data)

        # *******************************************************
        # TODO: Do things like checking correctness of frame here
        # and store it into a instance variable, e.g. self._frame
        # in case the frame is fragmented.
        # Below are some examples of how to do it
        # Have a look at the BMS base class for function to use,
        # take a look at other implementations for more  details
        # *******************************************************

        # if not data.startswith(BMS._HEAD):
        #     self._log.debug("incorrect SOF")
        #     return

        # if (crc := crc_sum(self._frame[:-1])) != self._frame[-1]:
        #     self._log.debug("invalid checksum 0x%X != 0x%X", self._frame[-1], crc)
        #     return

        # Do an immutable copy of the assembled (data) frame and notify _await_msg()
        # self._msg = bytes(self._frame)
        # self._msg_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        self._log.debug("replace with command to UUID %s", BMS.uuid_tx())
        # await self._await_msg(b"<some_command>")

        # TODO: parse data from self._frame here

        return {
            "voltage": 12,
            "current": 1.5,
            "temperature": 27.182,
        }  # TODO: fixed values, replace parsed data



"""Module to support RoyPow BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """RoyPow BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "RoyPow", "default_model": "smart BMS"}
    _HEAD: Final[bytes] = b"\xea\xd1\x01"
    _TAIL: Final[int] = 0xF5
    _BT_MODULE_MSG: Final[bytes] = b"AT+STAT\r\n"  # AT cmd from BLE module
    _MIN_LEN: Final[int] = len(_HEAD) + 1
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("battery_level", 7, 1, False, idx=0x4),
        BMSDp("voltage", 47, 2, False, lambda x: x / 100, 0x4),
        BMSDp(
            "current",
            6,
            3,
            False,
            lambda x: (x & 0xFFFF) * (-1 if (x >> 16) & 0x1 else 1) / 100,
            0x3,
        ),
        BMSDp("problem_code", 9, 3, False, idx=0x3),
        BMSDp(
            "cycle_charge",
            24,
            4,
            False,
            lambda x: ((x & 0xFFFF0000) | (x & 0xFF00) >> 8 | (x & 0xFF) << 8) / 1000,
            0x4,
        ),
        BMSDp("runtime", 30, 2, False, lambda x: x * 60, 0x4),
        BMSDp("temp_sensors", 13, 1, False, idx=0x3),
        BMSDp("cycles", 9, 2, False, idx=0x4),
        BMSDp("chrg_mosfet", 24, 1, False, lambda x: bool(x & 0x4), 0x3),
        BMSDp("dischrg_mosfet", 24, 1, False, lambda x: bool(x & 0x2), 0x3),
    )
    _CMDS: Final = frozenset({field.idx for field in _FIELDS}) | {0x2}

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: dict[int, bytes] = {}
        self._exp_len: int = 0

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "service_uuid": BMS.uuid_services()[0],
                "manufacturer_id": manufacturer_id,
                "connectable": True,
            }
            for manufacturer_id in (0x01A8, 0x0B31, 0x8AFB, 0x8849)
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe1"

    # async def _fetch_device_info(self) -> BMSInfo: unknown, use default

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        if not (data := data.removeprefix(BMS._BT_MODULE_MSG)):
            self._log.debug("filtering AT cmd")
            return

        if (
            data.startswith(BMS._HEAD)
            and not self._frame.startswith(BMS._HEAD)
            and len(data) > len(BMS._HEAD)
        ):
            self._exp_len = data[len(BMS._HEAD)]
            self._frame.clear()

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        if not self._frame.startswith(BMS._HEAD):
            self._frame.clear()
            return

        # verify that data is long enough
        if len(self._frame) < BMS._MIN_LEN + self._exp_len:
            return

        end_idx: Final[int] = BMS._MIN_LEN + self._exp_len - 1
        if self._frame[end_idx] != BMS._TAIL:
            self._log.debug("incorrect EOF: %s", self._frame)
            self._frame.clear()
            return

        if (crc := BMS._crc(self._frame[len(BMS._HEAD) : end_idx - 1])) != self._frame[
            end_idx - 1
        ]:
            self._log.debug(
                "invalid checksum 0x%X != 0x%X", self._frame[end_idx - 1], crc
            )
            self._frame.clear()
            return

        self._msg[self._frame[5]] = bytes(self._frame)
        self._frame.clear()
        self._msg_event.set()

    @staticmethod
    def _crc(frame: bytes | bytearray) -> int:
        """Calculate XOR of all frame bytes."""
        crc: int = 0
        for b in frame:
            crc ^= b
        return crc

    @staticmethod
    @cache
    def _cmd(cmd: bytes) -> bytes:
        """Assemble a RoyPow BMS command."""
        data: Final[bytes] = bytes([len(cmd) + 2, *cmd])
        return BMS._HEAD + data + bytes([BMS._crc(data), BMS._TAIL])

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        self._frame.clear()
        self._msg.clear()
        for cmd in BMS._CMDS:
            await self._await_msg(BMS._cmd(bytes([0xFF, cmd])))

        if not BMS._CMDS.issubset(self._msg.keys()):
            self._log.debug("Incomplete data set %s", self._msg.keys())
            raise ValueError("BMS data incomplete.")

        result: BMSSample = BMS._decode_data(BMS._FIELDS, self._msg)

        # remove remaining runtime if battery is charging
        if result.get("runtime") == 0xFFFF * 60:
            result.pop("runtime", None)

        result["cell_voltages"] = BMS._cell_voltages(
            self._msg.get(0x2, b""),
            cells=max(0, (len(self._msg.get(0x2, b"")) - 11) // 2),
            start=9,
        )
        result["temp_values"] = BMS._temp_values(
            self._msg.get(0x3, b""),
            values=result.get("temp_sensors", 0),
            start=14,
            size=1,
            signed=False,
            offset=40,
        )

        return result



"""Module to support EG4 BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus


class BMS(BaseBMS):
    """EG4 BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "EG4 electronics", "default_model": "LL"}
    _HEAD: Final[bytes] = b"\x01\x03"  # dev addr, fct code (read)
    _MAX_CELLS: Final[int] = 16
    _MAX_TEMP: Final[int] = 6
    _MIN_LEN: Final[int] = 5
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 3, 2, False, lambda x: x / 100),
        BMSDp("current", 5, 2, True, lambda x: x / 10),
        BMSDp("battery_health", 49, 2, False),
        BMSDp("battery_level", 51, 2, False),
        BMSDp("cycle_charge", 45, 2, False, lambda x: x / 10),
        BMSDp("cycles", 61, 4, False),
        BMSDp("cell_count", 75, 2, False),
        BMSDp("design_capacity", 77, 2, False, lambda x: x // 10),
        BMSDp("temperature", 39, 2, True),
        BMSDp("problem_code", 55, 6, False),
        BMSDp("balancer", 79, 2, False),
    )
    _OPT_FIELDS: Final[tuple[BMSValue, ...]] = (
        "cycle_charge",
        "cycles",
        "design_capacity",
    )

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""
        self._exp_len: int = 0

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "service_uuid": BMS.uuid_services()[0],
                "manufacturer_id": 0x6F80,
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> tuple[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("1000"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "1002"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "1001"

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if (
            len(data) > BMS._MIN_LEN
            and data.startswith(BMS._HEAD)
            and len(self._frame) >= self._exp_len
        ):
            self._exp_len = BMS._MIN_LEN + data[2]
            self._frame = bytearray()

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if len(self._frame) < self._exp_len:
            return

        del self._frame[self._exp_len :]

        if (crc := crc_modbus(self._frame[:-2])) != int.from_bytes(
            self._frame[-2:], byteorder="little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(self._frame[-2:], byteorder="little"),
                crc,
            )
            return

        self._msg = bytes(self._frame)
        self._msg_event.set()

    @staticmethod
    @cache
    def _cmd(address: int, count: int) -> bytes:
        """Assemble a EG4 BMS command."""
        frame: bytearray = bytearray(BMS._HEAD)
        frame += int.to_bytes(address, 2, byteorder="big")
        frame += int.to_bytes(count, 2, byteorder="big")
        frame += int.to_bytes(crc_modbus(frame), 2, byteorder="little")
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        await self._await_msg(BMS._cmd(0x0, 0x27))

        result: BMSSample = BMS._decode_data(BMS._FIELDS, self._msg)
        for field in BMS._OPT_FIELDS:
            if not result.get(field):
                del result[field]

        result["cell_voltages"] = BMS._cell_voltages(
            self._msg, cells=min(result.get("cell_count", 0), BMS._MAX_CELLS), start=7
        )
        result["temp_values"] = BMS._temp_values(
            self._msg, values=BMS._MAX_TEMP, start=69, size=1
        )
        return result



"""Module to support Jikong smart BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSMode, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str, crc_sum, lstr2int


class BMS(BaseBMS):
    """Jikong smart BMS class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Jikong", "default_model": "smart BMS"}
    _HEAD_RSP: Final = b"\x55\xaa\xeb\x90"  # header for responses
    _HEAD_CMD: Final = b"\xaa\x55\x90\xeb"  # cmd header (endianness!)
    _READY_MSG: Final = _HEAD_CMD + b"\xc8\x01\x01" + bytes(12) + b"\x44"
    _BT_MODULE_MSG: Final = b"\x41\x54\x0d\x0a"  # AT\r\n from BLE module
    _TYPE_POS: Final[int] = 4  # frame type is right after the header
    _INFO_LEN: Final[int] = 300
    _FIELDS: Final[tuple[BMSDp, ...]] = (  # Protocol: JK02_32S; JK02_24S has offset -32
        BMSDp("voltage", 150, 4, False, lambda x: x / 1000),
        BMSDp("current", 158, 4, True, lambda x: x / 1000),
        BMSDp("problem_code", 166, 4, False),
        BMSDp("balance_current", 170, 2, True, lambda x: x / 1000),
        BMSDp("balancer", 172, 1, False, bool),
        BMSDp("battery_level", 173, 1, False),
        BMSDp("cycle_charge", 174, 4, False, lambda x: x / 1000),
        BMSDp("design_capacity", 178, 4, False, lambda x: x // 1000),
        BMSDp("cycles", 182, 4, False),
        BMSDp("battery_health", 190, 1, False),
        BMSDp("chrg_mosfet", 198, 1, False, bool),
        BMSDp("dischrg_mosfet", 199, 1, False, bool),
        BMSDp("temp_sensors", 214, 2, True),
    )

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: bytes = b""
        self._char_write_handle: int = -1
        self._sw_version: int = 0
        self._prot_offset: int = 0
        self._valid_reply: int = 0x02
        self._bms_ready: bool = False

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
                "manufacturer_id": 0x0B65,
            },
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe1"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        self._valid_reply = 0x03
        await self._await_msg(self._cmd(b"\x97"), char=self._char_write_handle)
        self._valid_reply = 0x02
        return {
            "model": b2str(self._msg[6:22]),
            "hw_version": b2str(self._msg[22:30]),
            "sw_version": b2str(self._msg[30:38]),
            "name": b2str(self._msg[46:62]),
            "serial_number": b2str(self._msg[86:94]),
        }

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Retrieve BMS data update."""

        if data.startswith(BMS._BT_MODULE_MSG):
            self._log.debug("filtering AT cmd")
            if not (data := data.removeprefix(BMS._BT_MODULE_MSG)):
                return

        if (
            len(self._frame) >= self._INFO_LEN
            and (data.startswith((BMS._HEAD_RSP, BMS._HEAD_CMD)))
        ) or not self._frame.startswith(BMS._HEAD_RSP):
            self._frame.clear()

        self._frame += data

        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if (
            len(self._frame) < BMS._INFO_LEN and self._frame.startswith(BMS._HEAD_RSP)
        ) or len(self._frame) < BMS._TYPE_POS + 1:
            return

        # check that message type is expected
        if self._frame[BMS._TYPE_POS] != self._valid_reply:
            self._log.debug(
                "unexpected message type 0x%X (length %i): %s",
                self._frame[BMS._TYPE_POS],
                len(self._frame),
                self._frame,
            )
            return

        # trim AT\r\n message from the end
        if self._frame.endswith(BMS._BT_MODULE_MSG):
            self._log.debug("trimming AT cmd")
            self._frame = self._frame.removesuffix(BMS._BT_MODULE_MSG)

        # set BMS ready if msg is attached to last responses (v19.05)
        if self._frame[BMS._INFO_LEN :].startswith(BMS._READY_MSG):
            self._log.debug("BMS ready.")
            self._bms_ready = True
            del self._frame[BMS._INFO_LEN :]

        # trim message in case oversized
        if len(self._frame) > BMS._INFO_LEN:
            self._log.debug("wrong data length (%i): %s", len(self._frame), self._frame)
            del self._frame[BMS._INFO_LEN :]

        if (crc := crc_sum(self._frame[:-1])) != self._frame[-1]:
            self._log.debug("invalid checksum 0x%X != 0x%X", self._frame[-1], crc)
            return

        self._msg = bytes(self._frame)
        self._msg_event.set()

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        char_notify_handle: int = -1
        self._char_write_handle = -1
        self._bms_ready = False

        for service in self._client.services:
            self._log.debug("SRV %s", service)
            for char in service.characteristics:
                self._log.debug("  CHR %s: %s", char, ",".join(char.properties))
                if char.uuid == normalize_uuid_str(
                    BMS.uuid_rx()
                ) or char.uuid == normalize_uuid_str(BMS.uuid_tx()):
                    if "notify" in char.properties:
                        char_notify_handle = char.handle
                    if (
                        "write" in char.properties
                        or "write-without-response" in char.properties
                    ):
                        self._char_write_handle = char.handle
        if char_notify_handle == -1 or self._char_write_handle == -1:
            self._log.debug("failed to detect characteristics.")
            await self._client.disconnect()
            raise ConnectionError(f"Failed to detect characteristics from {self.name}.")
        self._log.debug(
            "using characteristics handle #%i (notify), #%i (write).",
            char_notify_handle,
            self._char_write_handle,
        )

        await super()._init_connection()

        # wait for BMS ready (0xC8)
        _bms_info: BMSInfo = await self._fetch_device_info()
        self._sw_version = lstr2int(_bms_info.get("sw_version", "0"))
        self._log.debug("device information: %s", _bms_info)
        self._prot_offset = -32 if self._sw_version < 11 else 0
        if not self._bms_ready:
            self._valid_reply = 0xC8  # BMS ready confirmation
            await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)
        self._valid_reply = 0x02  # cell information

    @staticmethod
    @cache
    def _cmd(cmd: bytes, value: list[int] | None = None) -> bytes:
        """Assemble a Jikong BMS command."""
        value = [] if value is None else value
        assert len(value) <= 13
        frame: bytearray = bytearray(
            [*BMS._HEAD_CMD, cmd[0], len(value), *value]
        ) + bytes(13 - len(value))
        frame.append(crc_sum(frame))
        return bytes(frame)

    def _temp_pos(self) -> list[tuple[int, int]]:
        if self._sw_version >= 14:
            return [(0, 144), (1, 162), (2, 164), (3, 254), (4, 256), (5, 258)]
        if self._sw_version >= 11:
            return [(0, 144), (1, 162), (2, 164), (3, 254)]
        return [(0, 130), (1, 132), (2, 134)]

    @staticmethod
    def _temp_sensors(
        data: bytes, temp_pos: list[tuple[int, int]], mask: int
    ) -> list[int | float]:
        return [
            (value / 10)
            for idx, pos in temp_pos
            if mask & (1 << idx)
            and (
                value := int.from_bytes(
                    data[pos : pos + 2], byteorder="little", signed=True
                )
            )
            != -2000
        ]

    @staticmethod
    def _conv_data(data: bytes, offs: int, sw_majv: int) -> BMSSample:
        """Return BMS data from status message."""

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, data, byteorder="little", start=offs
        )
        result["cell_count"] = int.from_bytes(
            data[70 + (offs >> 1) : 74 + (offs >> 1)], byteorder="little"
        ).bit_count()

        result["delta_voltage"] = (
            int.from_bytes(
                data[76 + (offs >> 1) : 78 + (offs >> 1)], byteorder="little"
            )
            / 1000
        )

        if sw_majv >= 15:
            result["battery_mode"] = (
                BMSMode(data[280 + offs])
                if data[280 + offs] in BMSMode
                else BMSMode.UNKNOWN
            )

        return result

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        if not self._msg_event.is_set() or self._msg[4] != 0x02:
            # request cell info (only if data is not constantly published)
            self._log.debug("requesting cell info")
            await self._await_msg(data=BMS._cmd(b"\x96"), char=self._char_write_handle)

        data: BMSSample = self._conv_data(
            self._msg, self._prot_offset, self._sw_version
        )
        data["temp_values"] = BMS._temp_sensors(
            self._msg, self._temp_pos(), data.get("temp_sensors", 0)
        )

        data["problem_code"] = (
            ((data.get("problem_code", 0)) >> 16)
            if self._prot_offset
            else (data.get("problem_code", 0) & 0xFFFF)
        )

        data["cell_voltages"] = BMS._cell_voltages(
            self._msg,
            cells=data.get("cell_count", 0),
            start=6,
            byteorder="little",
        )

        self._msg_event.clear()
        return data



"""Test the BLE Battery Management System base class functions."""

from collections.abc import Buffer, Callable
from logging import DEBUG
from string import hexdigits
from typing import Any, Final, Literal, NoReturn
from uuid import UUID

from bleak import BleakClient
from bleak.assigned_numbers import CharacteristicPropertyName
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection
from bleak.exc import BleakDeviceNotFoundError, BleakError
from bleak.uuids import normalize_uuid_str
from netaddr import OUI, NotRegisteredError
import pytest

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import (
    BaseBMS,
    b2str,
    crc8,
    crc_modbus,
    crc_sum,
    crc_xmodem,
    lrc_modbus,
)
from aiobmsble.bms.dummy_bms import BMS as DummyBMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient


class MockWriteModeBleakClient(MockBleakClient):
    """Emulate a BleakClient with selectable write mode response."""

    # The following attributes are used to simulate the behavior of the BleakClient
    # They need to be set via monkeypatching in the test since init() is called by the BMS
    PATTERN: list[bytes | Exception | None] = []  # data that is set to problem_code
    VALID_WRITE_MODES: list[CharacteristicPropertyName] = [
        "write-without-response",
        "write",
    ]
    EXP_WRITE_RESPONSE: list[bool] = []

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None
        if self.PATTERN:
            # check if we have a pattern to return
            pattern: bytes | Exception | None = self.PATTERN.pop(0)
            exp_wr_mode: Final[bool] = self.EXP_WRITE_RESPONSE.pop(0)
            if isinstance(pattern, Exception):
                raise pattern

            req_wr_mode: Final[str] = "write" if response else "write-without-response"
            assert response == exp_wr_mode, "write response mismatch"

            if isinstance(pattern, bytes) and req_wr_mode in self.VALID_WRITE_MODES:
                # check if we have a dict to return
                self._notify_callback("rx_char", bytearray(pattern))
                return

            # if None was selected do not return (trigger timeout) and wait for next pattern
            return

        # no pattern left, raise exception
        raise ValueError


class MinTestBMS(BaseBMS):
    """Minimal Test BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Test Manufacturer",
        "default_model": "minimal BMS for test",
    }

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"local_name": "Test", "connectable": True}]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("afe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "afe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "afe2"

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)
        # do not set event to make tests fail if wait_for_notify is not set

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_msg(b"mock_command", wait_for_notify=False)  # do not wait
        return {"problem_code": 21}


class DataTestBMS(MinTestBMS):
    """BMS providing simple data to test, e.g. value calculation."""

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset({"power", "battery_charging"})

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_msg(b"mock_command", wait_for_notify=False)  # do not wait
        return {
            "voltage": 13,
            "current": 1.7,
            "cycle_charge": 19,
            "cycles": 23,
            "problem_code": 21,
        }


class WMTestBMS(MinTestBMS):
    """Write mode mock BMS implementation."""

    def __init__(
        self,
        char_tx_properties: list[str],
        ble_device: BLEDevice,
        keep_alive: bool = True,
    ) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._char_tx_properties: list[str] = char_tx_properties

    def _wr_response(self, char: int | str) -> bool:
        return bool("write" in self._char_tx_properties)

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)
        self._data = data
        self._msg_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_msg(b"mock_command")

        return {"problem_code": int.from_bytes(self._data, "big", signed=False)}

class BMSBasicTests:
    """Base class for BMS tests."""
    bms_class: type[BaseBMS]

    def test_bms_id(self) -> None:
        """Test that the BMS returns default information."""

        for key in ("default_manufacturer", "default_model"):
            assert str(self.bms_class.INFO.get(key, "")).strip()
        assert len(self.bms_class.bms_id().strip())


    def test_matcher_dict(self) -> None:
        """Test that the BMS returns BT matcher."""

        assert len(self.bms_class.matcher_dict_list())
        for matcher in self.bms_class.matcher_dict_list():

            if manufacturer_id := matcher.get("manufacturer_id"):
                assert (
                    manufacturer_id == manufacturer_id & 0xFFFF
                ), f"incorrect {manufacturer_id=}"

            if service_uuid := matcher.get("service_uuid"):
                assert UUID(service_uuid), f"incorrect {service_uuid=}"

            if manufacturer_data_start := matcher.get("manufacturer_data_start"):
                assert all(
                    byte == byte & 0xFF for byte in manufacturer_data_start
                ), "manufacturer_data_start needs to contain Byte values!"

            if oui := matcher.get("oui"):
                parts: list[str] = oui.split(":")
                assert len(parts) == 3 and all(
                    len(part) == 2 and all(c in hexdigits for c in part) for part in parts
                ), f"incorrect {oui=}"
                try:
                    OUI(oui.replace(":", "-"))
                except NotRegisteredError:
                    assert (int(parts[0], 16) & 0xC0) not in (
                        0x00,  # Non-resolvable random private address
                        0x40,  # Resolvable random private address
                        # 0x80,  # 	Reserved for future use
                        # 0xC0,  # Static random device address
                    ), f"random private address OUI ({oui}) cannot be used for filtering!"

async def verify_device_info(
    patch_bleak_client,
    bleak_client: type[BleakClient],
    bms_class: type[BaseBMS],
    result_patch: dict[str, str] = {},
) -> None:
    """Test function for subclasses that the BMS returns device info from default characteristics."""
    patch_bleak_client(bleak_client)
    bms: BaseBMS = bms_class(generate_ble_device())
    assert (
        await bms.device_info()
        == {
            "fw_version": "mock_FW_version",
            "hw_version": "mock_HW_version",
            "sw_version": "mock_SW_version",
            "manufacturer": "mock_manufacturer",
            "model": "mock_model",
            "serial_number": "mock_serial_number",
        }
        | result_patch
    )


@pytest.mark.parametrize(
    ("bt_patch", "result_patch"),
    [
        ({"2a28": b"mock_SW_version"}, {"sw_version": "mock_SW_version"}),
        ({"2a28": b""}, {}),
        ({}, {}),
    ],
    ids=["defaults", "empty", "no_char"],
)
async def test_device_info(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    bt_patch: dict[str, bytes],
    result_patch: dict[str, str],
) -> None:
    """Verify that device_info reads BLE characteristic 180A and provides default values."""
    defaults: dict[str, bytes] = MockBleakClient.BT_INFO.copy()
    del defaults["2a28"]
    monkeypatch.setattr(MockBleakClient, "BT_INFO", defaults | bt_patch)
    patch_bleak_client()
    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    assert (
        await bms.device_info()
        == {
            "fw_version": "mock_FW_version",
            "model": "mock_model",
            "serial_number": "mock_serial_number",
            "hw_version": "mock_HW_version",
            "manufacturer": "mock_manufacturer",
        }
        | result_patch
    )


async def test_device_info_fail(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client: Callable[..., None]
) -> None:
    """Test only BMS default information is returned if characteristic 0x180a does not exit."""
    monkeypatch.setattr(
        BleakGATTServiceCollection, "get_service", lambda obj, char: None
    )
    patch_bleak_client()
    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    await bms.async_update()  # run update to have connection open
    assert not await bms.device_info()  # if characteristic does not exist, no output
    assert bms.name == "MockBLEDevice"  # name is gathered from BLEDevice
    assert bms._client.is_connected


def test_calc_pwr_chrg_temp(bms_data_fixture: BMSSample) -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = bms_data_fixture
    ref: BMSSample = bms_data_fixture.copy()

    BaseBMS._add_missing_values(bms_data)
    ref = ref | {
        "cell_count": 2,
        "cycle_capacity": 238,
        "delta_voltage": 0.111,
        "power": (
            -91
            if bms_data.get("current", 0) < 0
            else 0 if bms_data.get("current") == 0 else 147
        ),
        # battery is charging if current is positive
        "battery_charging": bms_data.get("current", 0) > 0,
        "temperature": -34.396,
        "problem": False,
    }
    if bms_data.get("current", 0) < 0:
        ref |= {"runtime": 9415}

    assert bms_data == ref


@pytest.mark.parametrize(
    ("sample", "expected"),
    [
        (
            {"cell_voltages": [3.456, 3.567]},
            {"cell_count": 2, "delta_voltage": 0.111, "voltage": 7.023},
        ),
        ({"battery_level": 73, "design_capacity": 125}, {"cycle_charge": 91.25}),
        ({"cycle_charge": 421, "design_capacity": 983}, {"battery_level": 42.8}),
        ({"total_charge": 1234567, "design_capacity": 256}, {"cycles": 4822}),
        (
            {"current": -1.3, "cycle_charge": 73},
            {"battery_charging": False, "runtime": 202153},
        ),
        ({"current": 1.3, "cycle_charge": 73}, {"battery_charging": True}),
    ],
    ids=["voltage", "cycle_charge", "battery_level", "cycles", "runtime", "no_runtime"],
)
def test_calc_values(sample: BMSSample, expected: BMSSample) -> None:
    """Check if missing data is correctly calculated."""
    ref: BMSSample = sample.copy()
    BaseBMS._add_missing_values(sample)
    assert sample == ref | expected | {"problem": False}


@pytest.mark.parametrize(
    ("raw"),
    [True, False],
    ids=["raw", "ext"],
)
async def test_async_update(patch_bleak_client: Callable[..., None], raw: bool) -> None:
    """Check update function of the BMS returns values."""
    patch_bleak_client()
    bms: DataTestBMS = DataTestBMS(generate_ble_device())
    base_result: BMSSample = BMSSample(
        {
            "voltage": 13,
            "current": 1.7,
            "cycle_charge": 19,
            "cycles": 23,
            "problem_code": 21,
        }
    )
    if not raw:
        base_result.update(
            {
                "battery_charging": True,
                "cycle_capacity": 247,
                "power": 22.1,
                "problem": True,
            }
        )
    assert await bms.async_update(raw=raw) == base_result


@pytest.mark.parametrize(
    ("problem_sample"),
    [
        ({"voltage": -1}, "negative overall voltage"),
        ({"cell_voltages": [5.907]}, "high cell voltage"),
        ({"cell_voltages": [-0.001]}, "negative cell voltage"),
        ({"delta_voltage": 5.907}, "doubtful delta voltage"),
        ({"cycle_charge": 0}, "doubtful cycle charge"),
        ({"battery_level": 101}, "doubtful SoC"),
        ({"problem_code": 0x1}, "BMS problem code"),
        ({"problem": True}, "BMS problem report"),
    ],
    ids=lambda param: param[1],
)
def test_problems(problem_sample: tuple[BMSSample, str]) -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = problem_sample[0].copy()

    BaseBMS._add_missing_values(bms_data)

    assert ("problem", True) in bms_data.items()


@pytest.mark.parametrize(
    ("replies", "exp_wr_response", "exp_output"),
    [
        ([b"\x12"], [True], [0x12]),
        (
            [None] * 2 * (BaseBMS.MAX_RETRY),
            [True] * (BaseBMS.MAX_RETRY) + [False] * (BaseBMS.MAX_RETRY),
            [TimeoutError()],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY - 1) + [b"\x13"],
            [True] * (BaseBMS.MAX_RETRY),
            [0x13],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY) + [b"\x14"],
            [True] * (BaseBMS.MAX_RETRY) + [False],
            [0x14],
        ),
        (
            [BleakError()]
            + [None] * (BaseBMS.MAX_RETRY - 1)
            + [b"\x15"]
            + [None] * (BaseBMS.MAX_RETRY - 1)
            + [b"\x16"],
            [True] + [False] * BaseBMS.MAX_RETRY + [False] * BaseBMS.MAX_RETRY,
            [0x15, 0x16],
        ),
        (
            [BleakDeviceNotFoundError("mock_device")],
            [False],
            [BleakDeviceNotFoundError("mock_device")],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY - 1) + [ValueError()],
            [True] * (BaseBMS.MAX_RETRY),
            [ValueError()],
        ),
    ],
    ids=[
        "basic_test",
        "no_response",
        "retry_count-1",
        "retry_count",
        "mode_switch",
        "no-retry-exc",
        "unhandled_exc",
    ],
)
async def test_write_mode(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    patch_bms_timeout: Callable[..., None],
    replies: list[bytearray | Exception | None],
    exp_wr_response: list[bool],
    exp_output: list[int | Exception],
    request: pytest.FixtureRequest,
) -> None:
    """Check if write mode selection works correctly."""

    assert len(replies) == len(
        exp_wr_response
    ), "Replies and expected responses must match in length!"
    patch_bms_timeout()
    monkeypatch.setattr(MockWriteModeBleakClient, "PATTERN", replies)
    monkeypatch.setattr(MockWriteModeBleakClient, "EXP_WRITE_RESPONSE", exp_wr_response)

    patch_bleak_client(MockWriteModeBleakClient)

    bms = WMTestBMS(
        ["write-no-response", "write"],
        generate_ble_device(),
        False,
    )

    # NOTE: output must reflect the end result after one call, as init of HA resets the whole BMS!
    for output in exp_output:
        if isinstance(output, Exception):
            with pytest.raises(type(output)):
                await bms.async_update()
        else:
            assert await bms.async_update() == {
                "problem": (output != 0),
                "problem_code": output,
            }, f"{request.node.name} failed!"


async def test_wr_mode_reset(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client: Callable[..., None]
) -> None:
    """Check that write mode selection is reset on disconnect of the BMS."""
    monkeypatch.setattr(MockWriteModeBleakClient, "PATTERN", [b"\x42"])
    monkeypatch.setattr(MockWriteModeBleakClient, "EXP_WRITE_RESPONSE", [False])
    patch_bleak_client(MockWriteModeBleakClient)

    bms: WMTestBMS = WMTestBMS(["write_without"], generate_ble_device())
    assert await bms.async_update() == {"problem": True, "problem_code": 0x42}
    assert bms._inv_wr_mode is False
    await bms.disconnect(True)
    assert bms._inv_wr_mode is None


def test_get_bms_module() -> None:
    """Check that basebms and dummy_bms return correct module name."""
    assert BaseBMS.get_bms_module() == "aiobmsble.basebms"
    assert DummyBMS.get_bms_module() == "aiobmsble.bms.dummy_bms"


async def test_no_notify(
    patch_bleak_client: Callable[..., None], caplog: pytest.LogCaptureFixture
) -> None:
    """Test BMS update without waiting for notification event."""
    patch_bleak_client(MockBleakClient)

    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
    with caplog.at_level(DEBUG):
        result: BMSSample = await bms.async_update()
    assert "MockBleakClient write_gatt_char afe2, data: b'mock_command'" in caplog.text
    assert result == {"problem": True, "problem_code": 21}
    assert not bms._client.is_connected


async def test_disconnect_fail(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check that exceptions in connect function for guarding disconnect are ignored."""

    async def _raise_bleak_error(*args: Any) -> NoReturn:
        raise BleakError

    monkeypatch.setattr(MockBleakClient, "disconnect", _raise_bleak_error)
    patch_bleak_client(MockBleakClient)

    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
    with caplog.at_level(DEBUG):
        result: BMSSample = await bms.async_update()
    assert result == {"problem": True, "problem_code": 21}
    assert "failed to disconnect stale connection (BleakError)" in caplog.text
    assert "disconnect failed!" in caplog.text


async def test_init_connect_fail(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check that exceptions in connect function for guarding disconnect are ignored."""

    async def _raise_value_error(*args: Any) -> NoReturn:
        raise ValueError("MockValueError")

    patch_bleak_client(MockBleakClient)
    monkeypatch.setattr(MinTestBMS, "_init_connection", _raise_value_error)

    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    with caplog.at_level(DEBUG), pytest.raises(ValueError, match="MockValueError"):
        await bms.async_update()


async def test_context_mgr(
    patch_bleak_client: Callable[..., None],
) -> None:
    """Test that context manager provides data."""
    patch_bleak_client(MockBleakClient)

    async with DataTestBMS(generate_ble_device(), keep_alive=True) as bms:
        assert await bms.async_update() == {
            "voltage": 13,
            "current": 1.7,
            "battery_charging": True,
            "power": 22.1,
            "cycle_capacity": 247,
            "cycle_charge": 19,
            "cycles": 23,
            "problem": True,
            "problem_code": 21,
        }


async def test_context_mgr_fail(
    patch_bleak_client: Callable[..., None],
) -> None:
    """Check that context manager enforces `keep_alive=True`."""

    patch_bleak_client(MockBleakClient)

    with pytest.raises(ValueError, match="usage of context manager*"):
        async with MinTestBMS(generate_ble_device(), keep_alive=False) as bms:
            await bms.async_update()


def test_crc_calculations() -> None:
    """Check if CRC calculations are correct."""
    # Example data for CRC calculation
    data: bytearray = bytearray([0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39])
    test_fn: list[tuple[Callable[[bytearray], int], int]] = [
        (crc_modbus, 0x4B37),
        (crc8, 0xA1),
        (crc_xmodem, 0x31C3),
        (lrc_modbus, 0xFE23),
        (crc_sum, 0xDD),
    ]

    for crc_fn, expected_crc in test_fn:
        calculated_crc: int = crc_fn(data)
        assert (
            calculated_crc == expected_crc
        ), f"Expected {expected_crc}, got {calculated_crc}"


@pytest.mark.parametrize(
    ("data", "cells", "start", "size", "byteorder", "divider", "expected"),
    [
        # Two cells, big endian, default divider
        (b"\x0d\x80\x0d\xf7", 2, 0, 2, "big", 1000, [3.456, 3.575]),
        # Two cells, little endian, default divider
        (b"\x80\x0d\xf7\x0d", 2, 0, 2, "little", 1000, [3.456, 3.575]),
        # One cell, big endian, custom divider
        (b"\x30\x34", 1, 0, 2, "big", 10000, [1.234]),
        # Three cells, big endian, offset start
        (
            b"\x00\x00\x0d\x80\x0d\xf7\x0d\xa7",
            3,
            2,
            2,
            "big",
            1000,
            [3.456, 3.575, 3.495],
        ),
        # Not enough data for all cells
        (b"\x0d\x80", 2, 0, 2, "big", 1000, [3.456]),
        # Zero cells
        (b"\x0d\x80", 0, 0, 2, "big", 1000, []),
        # Divider = 1 (raw values)
        (b"\x01\x02\x03\x04", 2, 0, 2, "big", 1, [258, 772]),
    ],
    ids=[
        "two_cells_big_endian",
        "two_cells_little_endian",
        "one_cell_custom_divider",
        "three_cells_offset_start",
        "not_enough_data",
        "zero_cells",
        "divider_one_raw_values",
    ],
)
def test_cell_voltages(
    data: bytes,
    cells: int,
    start: int,
    size: int,
    byteorder: Literal["little", "big"],
    divider: int,
    expected: list[float],
) -> None:
    """Test the _cell_voltages method of BaseBMS with various input parameters."""
    result: list[float] = BaseBMS._cell_voltages(
        data,
        cells=cells,
        start=start,
        size=size,
        byteorder=byteorder,
        divider=divider,
    )
    assert result == expected


@pytest.mark.parametrize(
    (
        "data",
        "values",
        "start",
        "size",
        "byteorder",
        "signed",
        "offset",
        "divider",
        "expected",
    ),
    [
        # Two signed big endian values, no offset, divider=1
        (b"\xff\xec\x00\x64", 2, 0, 2, "big", True, 0, 1, [-20, 100]),
        # Two unsigned little endian values, offset=0, divider=10
        (b"\x10\x00\x20\x00", 2, 0, 2, "little", False, 0, 10, [1.6, 3.2]),
        # One signed big endian value, offset=273.15, divider=1
        (b"\x0b\x98", 1, 0, 2, "big", False, 2731, 10, [23.7]),
        # Three signed little endian values, offset=0, divider=100
        (
            b"\x64\x00\xc8\xff\x2c\x01",
            3,
            0,
            2,
            "little",
            True,
            0,
            100,
            [1.0, -0.56, 3.0],
        ),
        # Not enough data for all values
        (b"\x00\x7d", 2, 0, 2, "big", True, 0, 1, [125]),
        # Zero values requested
        (b"\x00\x7d", 0, 0, 2, "big", True, 0, 1, []),
        # Divider = 1, offset = 7
        (b"\x00\x14", 1, 0, 2, "big", True, 7, 1, [13]),
        # no offset, div = 10
        (b"\x65\x64\x00\x40", 4, 0, 1, "big", True, 0, 10, [10.1, 10.0, 0.0, 6.4]),
        # offset -40, div = 10
        (b"\x65\x64\x00\x40", 4, 0, 1, "big", True, 40, 10, [6.1, 6.0, 2.4]),
        # offset -25, div = 1
        (b"\x65\x64\x00\x40", 4, 0, 1, "big", True, 25, 1, [76, 75, 39]),
    ],
    ids=[
        "two_signed_big_endian",
        "two_unsigned_little_endian_div10",
        "one_signed_big_endian_kelvin_offset",
        "three_signed_little_endian_div100",
        "not_enough_data",
        "zero_values",
        "divider1_offset10",
        "no offset",
        "offset-40",
        "div-1",
    ],
)
def test_temp_values(
    data: bytes,
    values: int,
    start: int,
    size: int,
    byteorder: Literal["little", "big"],
    signed: bool,
    offset: float,
    divider: int,
    expected: list[int | float],
) -> None:
    """Test the _temp_values method of BaseBMS with various input parameters."""
    result: list[int | float] = BaseBMS._temp_values(
        data,
        values=values,
        start=start,
        size=size,
        byteorder=byteorder,
        signed=signed,
        offset=offset,
        divider=divider,
    )
    assert result == expected


@pytest.mark.parametrize(
    ("fields", "data", "byteorder", "start", "expected"),
    [
        # Test with big endian and multiple data points
        (
            (
                BMSDp("voltage", 0, size=2, signed=False),
                BMSDp("current", 2, size=2, signed=True),
            ),
            b"\x0d\x80\xff\xec",
            "big",
            0,
            {"voltage": 3456.0, "current": -20},
        ),
        # Test with dict data, little endian
        (
            (BMSDp("voltage", 0, size=2, signed=False, fct=lambda x: x / 10, idx=1),),
            {1: b"\x64\x00"},
            "little",
            0,
            {"voltage": 10},
        ),
        # Test with missing dict data
        (
            (BMSDp("voltage", 0, size=2, signed=False, idx=2),),
            {1: b"\x64\x00"},
            "little",
            0,
            {},
        ),
        # Test with start
        (
            (BMSDp("battery_level", 1, size=1, signed=False),),
            b"\x00\x7d",
            "big",
            0,
            {"battery_level": 125},
        ),
        # Test with start shifting the slice
        (
            (BMSDp("voltage", 0, size=2, signed=False),),
            b"\x00\x00\x12\x34",
            "big",
            2,
            {"voltage": 0x1234},
        ),
    ],
    ids=[
        "be_multiple_dps",
        "le_dict_single_dp_lambda",
        "le_dict_missing_idx",
        "be_with_start",
        "be_start_shift_slice",
    ],
)
def test_decode_data(
    fields: tuple[BMSDp, ...],
    data: bytes | dict[int, bytes],
    byteorder: Literal["little", "big"],
    start: int,
    expected: BMSSample,
) -> None:
    """Test the _decode_data method of BaseBMS with various input parameters."""
    result: BMSSample = BaseBMS._decode_data(
        fields, data, byteorder=byteorder, start=start
    )
    assert result == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [(b"", ""), (b"\x00 ", ""), (b"test\x00 ", "test"), (b"test  \t\r ", "test")],
    ids=["empty", "hex", "text_hex", "test_space"],
)
def test_b2str(data: bytes, expected: str) -> None:
    """Test bytearray to string conversion function."""
    assert b2str(data) == expected



"""Module to support AG Automotive (E&J) BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from aiobmsble import BMSInfo, MatcherPattern
from aiobmsble.bms.ej_bms import BMS as EJBMS


class BMS(EJBMS):
    """AG Automotive (E&J) BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "AG Automotive",
        "default_model": "AG Power Lithium",
    }

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            MatcherPattern(
                local_name="AG???L*",
                manufacturer_id=21320,
                connectable=True,
            )
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return ("00008000-0000-1000-8000-57616c6b697a",)

    @staticmethod
    def uuid_rx() -> str:
        """Return 128-bit UUID of characteristic that provides notification/read property."""
        return "00008002-0000-1000-8000-57616c6b697a"

    @staticmethod
    def uuid_tx() -> str:
        """Return 128-bit UUID of characteristic that provides write property."""
        return "00008001-0000-1000-8000-57616c6b697a"



"""Module to support Lithionics BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Lithionics BMS implementation (ASCII stream protocol)."""

    INFO: BMSInfo = {
        "default_manufacturer": "Lithionics",
        "default_model": "NeverDie smart BMS",
    }
    _HEAD_STATUS: Final[str] = "&,"
    _MIN_FIELDS_PRIMARY: Final[int] = 10
    _MIN_FIELDS_STATUS: Final[int] = 3

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._stream_data: dict[str, list[str]] = {}

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            # Seen on Lithionics Li3 packs: "Li3-061322094"
            MatcherPattern(
                local_name="Li[0-9]-*",
                service_uuid=BMS.uuid_services()[0],
                manufacturer_id=19784,
                connectable=True,
            ),
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        raise NotImplementedError

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        self._frame += data
        while (idx := self._frame.find(b"\r\n")) >= 0:
            line: str = self._frame[:idx].decode("ascii", errors="ignore").strip()
            del self._frame[: idx + 2]

            if not line:
                continue

            if line == "ERROR":
                self._log.debug("ignoring command response: %s", line)
                continue

            fields: list[str] = line.split(",")
            if (
                line.startswith(BMS._HEAD_STATUS)
                and len(fields) >= BMS._MIN_FIELDS_STATUS
            ):
                self._stream_data["status"] = fields
            elif line[0].isdigit() and len(fields) >= BMS._MIN_FIELDS_PRIMARY:
                self._stream_data["primary"] = fields

            if self._stream_data.keys() >= {"primary", "status"}:
                self._msg_event.set()

    @staticmethod
    def _parse_primary(fields: list[str]) -> BMSSample:
        # BMS reports temperatures in Fahrenheit.
        temp_values: Final[list[float]] = [
            round((int(fields[idx]) - 32) * 5 / 9, 3) for idx in (5, 6)
        ]

        return {
            "voltage": int(fields[0]) / 100,
            "cell_voltages": [int(value) / 100 for value in fields[1:5]],
            "temp_values": temp_values,
            "temp_sensors": 2,
            "current": float(fields[7]),
            "battery_level": int(fields[8]),
            "problem_code": int(fields[9], 16),
        }

    @staticmethod
    def _parse_status(fields: list[str]) -> BMSSample:

        result: BMSSample = {"cycle_charge": float(fields[2])}
        if len(fields) > 3:
            result["total_charge"] = int(fields[3])

        return result

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        self._stream_data.clear()
        self._msg_event.clear()
        await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)

        result: BMSSample = BMS._parse_primary(
            self._stream_data["primary"]
        ) | BMS._parse_status(self._stream_data["status"])

        return result



"""Test the JBD BMS implementation."""

import asyncio
from collections.abc import Buffer, Callable, Iterable
from typing import Any, Final
from uuid import UUID

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.jbd_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 20

_RESULT_DEFS: Final[BMSSample] = {
    "temp_sensors": 3,
    "voltage": 15.6,
    "current": -2.87,
    "battery_level": 100,
    "cycle_charge": 4.98,
    "design_capacity": 5,
    "cycles": 42,
    "temperature": 22.133,
    "cycle_capacity": 77.688,
    "power": -44.772,
    "battery_charging": False,
    "runtime": 6246,
    "cell_count": 4,
    "cell_voltages": [3.43, 3.425, 3.432, 3.417],
    "temp_values": [22.4, 22.3, 21.7],
    "delta_voltage": 0.015,
    "problem": False,
    "problem_code": 0,
    "balancer": 0,
    "chrg_mosfet": True,
    "dischrg_mosfet": True,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockJBDBleakClient(MockBleakClient):
    """Emulate a JBD BMS BleakClient."""

    HEAD_CMD = 0xDD
    CMD_INFO = bytearray(b"\xa5\x03")
    CMD_CELL = bytearray(b"\xa5\x04")
    HW_INFO = bytearray(b"\xa5\x05")
    ACK_MSG = bytearray(b"\xff\xaa\x15\x01\x00\x16")
    REQUIRE_PASS = False
    UNLOCKED = False
    DEFAULT_SECRET = b"000000"

    _tasks: set[asyncio.Task[None]] = set()

    def __init__(
        self,
        address_or_ble_device: BLEDevice,
        disconnected_callback: Callable[[BleakClient], None] | None,
        services: Iterable[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize MockBleakClient."""
        super().__init__(
            address_or_ble_device, disconnected_callback, services, **kwargs
        )
        self._services = ["ff01", "ff02"]

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:

        _msg: Final[bytes] = bytes(data)
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("ff02")
            and _msg.startswith(b"\xff\xaa\x15")
            and not self.UNLOCKED
        ):
            self.UNLOCKED = True
            return (
                MockJBDBleakClient.ACK_MSG
                if sum(_msg[2:-1]) & 0xFF == _msg[-1]
                and _msg[4:-1] == self.DEFAULT_SECRET
                else bytearray(b"\xff\xaa\x15\x01\x01\x17")
            )

        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("ff02")
            and _msg[0] == self.HEAD_CMD
            and self.REQUIRE_PASS == self.UNLOCKED
        ):
            match _msg[1:3]:
                case self.CMD_INFO:
                    return bytearray(
                        b"\xdd\x03\x00\x1d\x06\x18\xfe\xe1\x01\xf2\x01\xf4\x00\x2a\x2c\x7c\x00\x00\x00"
                        b"\x00\x00\x00\x80\x64\x03\x04\x03\x0b\x8b\x0b\x8a\x0b\x84\xf8\x84\x77"
                    )  # {'voltage': 15.6, 'current': -2.87, 'battery_level': 100, 'cycle_charge': 4.98, 'cycles': 42, 'temperature': 22.133333333333347}
                case self.CMD_CELL:
                    return bytearray(
                        b"\xdd\x04\x00\x08\x0d\x66\x0d\x61\x0d\x68\x0d\x59\xfe\x3c\x77"
                    )  # {'cell#0': 3.43, 'cell#1': 3.425, 'cell#2': 3.432, 'cell#3': 3.417}
                case self.HW_INFO:
                    return bytearray(
                        b"\xdd\x05\x00\x0a\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\xfd\xe9\x77"
                    )  # hardware version 0123456789
        return bytearray()

    async def _send_data(self, char_specifier, data) -> None:
        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        # always send two responses, to test timeout behaviour
        for resp in (
            self._response(char_specifier, bytearray(b"\xdd\xa5\x03\x00\xff\xfd\x77")),
            self._response(char_specifier, data),
        ):
            for notify_data in [
                resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
            ]:
                self._notify_callback("MockJBDBleakClient", notify_data)
            await asyncio.sleep(0)

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        _task: asyncio.Task[None] = asyncio.create_task(
            self._send_data(char_specifier, data)
        )
        self._tasks.add(_task)
        _task.add_done_callback(self._tasks.discard)

    async def disconnect(self) -> None:
        """Mock disconnect."""
        await asyncio.gather(*self._tasks)
        await super().disconnect()


class MockOversizedBleakClient(MockJBDBleakClient):
    """Emulate a JBD BMS BleakClient returning wrong data length."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("ff02")
            and bytearray(data)[0] == self.HEAD_CMD
        ):
            if bytearray(data)[1:3] == self.CMD_INFO:
                return bytearray(
                    b"\xdd\x03\x00\x1d\x06\x18\xfe\xe1\x01\xf2\x01\xf4\x00\x2a\x2c\x7c\x00\x00\x00"
                    b"\x00\x00\x00\x80\x64\x03\x04\x03\x0b\x8b\x0b\x8a\x0b\x84\xf8\x84\x77"
                    b"\00\00\00\00\00\00"  # oversized response
                )  # {'voltage': 15.6, 'current': -2.87, 'battery_level': 100, 'cycle_charge': 4.98, 'cycles': 42, 'temperature': 22.133333333333347}
            if bytearray(data)[1:3] == self.CMD_CELL:
                return bytearray(
                    b"\xdd\x04\x00\x08\x0d\x66\x0d\x61\x0d\x68\x0d\x59\xfe\x3c\x77"
                    b"\00\00\00\00\00\00\00\00\00\00\00\00"  # oversized response
                )  # {'cell#0': 3.43, 'cell#1': 3.425, 'cell#2': 3.432, 'cell#3': 3.417}

        return bytearray()

    async def disconnect(self) -> None:
        """Mock disconnect to raise BleakError."""
        if self._tasks:
            await asyncio.wait(self._tasks)
        raise BleakError


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test JBD BMS data update."""

    patch_bleak_client(MockJBDBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.mark.parametrize(
    "secret", ["000000", "wrong"], ids=["correct_secret", "wrong_secret"]
)
async def test_update_secret(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, secret: str
) -> None:
    """Test JBD BMS data update."""

    monkeypatch.setattr(MockJBDBleakClient, "REQUIRE_PASS", True)
    patch_bleak_client(MockJBDBleakClient)

    bms = BMS(generate_ble_device(), secret=secret)
    if secret == "wrong":
        with pytest.raises(PermissionError):
            await bms.async_update()
    else:
        assert await bms.async_update() == _RESULT_DEFS

    await bms.disconnect()


@pytest.mark.parametrize(
    "wrong_init",
    [
        b"",
        b"\xf0\xaa\x15\x01\x00\x16",
        b"\xff\xaa\x15\x01\x16",
        b"\xff\xaa\x15\x01\x00\x15",
        b"\xff\xaa\x16\x01\x00\x17",
    ],
    ids=["empty_resp", "wrong_SOF", "wrong_len", "wrong_CRC", "wrong_cmd"],
)
async def test_invalid_init(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_init: bytes,
) -> None:
    """Test connection init with BMS returning invalid data (wrong CRC)."""

    patch_bms_timeout()
    monkeypatch.setattr(MockJBDBleakClient, "REQUIRE_PASS", True)
    monkeypatch.setattr(MockJBDBleakClient, "ACK_MSG", bytearray(wrong_init))
    patch_bleak_client(MockJBDBleakClient)

    bms = BMS(generate_ble_device(), secret="000000")

    with pytest.raises(TimeoutError):
        _result: BMSSample = await bms.async_update()

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockJBDBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {"hw_version": "0123456789"}


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            bytearray(
                b"\xdd\x03\x00\x1d\x06\x18\xfe\xe1\x01\xf2\x01\xf4\x00\x2a\x2c\x7c\x00\x00\x00"
                b"\x00\x00\x00\x80\x64\x03\x04\x03\x0b\x8b\x0b\x8a\x0b\x84\xf8\x84\xdd"
            ),
            "wrong end",
        ),
        (bytearray(b"\xdd\x04\x00\x1d" + b"\x00" * 31 + b"\x77"), "wrong CRC"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data update with BMS returning invalid data (wrong CRC)."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockJBDBleakClient, "_response", lambda _s, _c, _d: wrong_response
    )

    patch_bleak_client(MockJBDBleakClient)

    bms = BMS(generate_ble_device())

    with pytest.raises(TimeoutError):
        _result: BMSSample = await bms.async_update()

    await bms.disconnect()


async def test_oversized_response(patch_bleak_client) -> None:
    """Test data update with BMS returning oversized data, result shall still be ok."""
    patch_bleak_client(MockOversizedBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.async_update() == _RESULT_DEFS
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\xdd\x03\x00\x1d\x06\x18\xfe\xe1\x01\xf2\x01\xf4\x00\x2a\x2c\x7c\x00\x00\x00"
                b"\x00\x00\x01\x80\x64\x03\x04\x03\x0b\x8b\x0b\x8a\x0b\x84\xf8\x83\x77"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\xdd\x03\x00\x1d\x06\x18\xfe\xe1\x01\xf2\x01\xf4\x00\x2a\x2c\x7c\x00\x00\x00"
                b"\x00\x80\x00\x80\x64\x03\x04\x03\x0b\x8b\x0b\x8a\x0b\x84\xf8\x04\x77"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning invalid data (wrong CRC)."""

    def _response(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        resp: bytearray = problem_response[0],
    ) -> bytearray:
        if (
            isinstance(char_specifier, str)
            and normalize_uuid_str(char_specifier) == normalize_uuid_str("ff02")
            and bytearray(data)[0] == self.HEAD_CMD
        ):
            if bytearray(data)[1:3] == self.CMD_INFO:
                return resp
            if bytearray(data)[1:3] == self.CMD_CELL:
                return bytearray(
                    b"\xdd\x04\x00\x08\x0d\x66\x0d\x61\x0d\x68\x0d\x59\xfe\x3c\x77"
                )  # {'cell#0': 3.43, 'cell#1': 3.425, 'cell#2': 3.432, 'cell#3': 3.417}

        return bytearray()

    monkeypatch.setattr(MockJBDBleakClient, "_response", _response)
    patch_bleak_client(MockJBDBleakClient)
    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 15),
    }

    await bms.disconnect()
